/*
 * File: \data-structures\linked-list.js                                       *
 */

class ListNode {
  constructor(value = null) {
    this.next = null;
    this.value = value;
  }
}

class LinkedList {
  constructor() {
    this.head = null;
    this.tail = null;
    this.elements = 0;
  }

  /**
   * Inserts a new value (node) in the LinkedList
   * @params value:Number. Value to be inserted in the LinkedList
   * @returns boolean. If value is correctly inserted returns true, otherwise false.
   */
  insert(value = null) {
    if (value === null) return false;

    const newNode = new ListNode(value);
    this.elements++;

    if (!this.head) {
      this.head = this.tail = newNode;
      return true;
    }

    this.tail.next = newNode;
    this.tail = newNode;
    return true;
  }

  /**
   * Deletes the node at position passed by paramenter from the LinkedList
   * @params pos:Number. Position of the LinkedList to delete. Starts from 0.
   * @returns boolean. If node is deleted returns true, otherwise false.
   */
  deleteByPosition(pos = null) {
    // Input validation
    if (pos === null || pos < 0 || pos >= this.elements) return false;

    // Special case: deleting head
    if (pos === 0) {
      this.head = this.head.next;
      this.elements--;
      // Update tail if list becomes empty
      if (!this.head) this.tail = null;
      return true;
    }

    // Traverse to the node before the one we want to delete
    let current = this.head;
    for (let i = 0; i < pos - 1 && current; i++) {
      current = current.next;
    }

    // Safety check in case of corrupted list
    if (!current || !current.next) return false;

    // Update references to skip the node we're deleting
    current.next = current.next.next;

    // Update tail if we're deleting the last node
    if (pos === this.elements - 1) this.tail = current;

    this.elements--;
    return true;
  }

  /**
   * Deletes the first node with value passed by paramenter from the LinkedList
   * @params value:Number. Value of the first node encountered to delete from the LinkedList.
   * @returns boolean. If a node is deleted returns true, otherwise false.
   */
  deleteByValue(value = null) {
    if (value === null || !this.head) return false;

    // Special case: deleting head
    if (this.head.value === value) {
      this.head = this.head.next;
      this.elements--;
      if (!this.head) this.tail = null;
      return true;
    }

    // Use two pointers for cleaner traversal
    let prev = this.head;
    let current = this.head.next;

    while (current) {
      if (current.value === value) {
        prev.next = current.next;
        if (!prev.next) this.tail = prev;
        this.elements--;
        return true;
      }
      prev = current;
      current = current.next;
    }

    return false;
  }

  /**
   * Searches the value by position passed as paramenter in the LinkedList.
   * @params pos:Number. Position to extract value from the LinkedList.
   * @returns Number. If the position is between limits returns the value associated with the node in such position, otherwise returns false
   */
  searchByPosition(pos = null) {
    if (pos === null || pos < 0 || pos >= this.elements) return false;

    let current = this.head;
    for (let i = 0; i < pos; i++) {
      current = current.next;
    }

    return current.value;
  }

  /**
   * Searches a node with value passed as paramenter in the LinkedList.
   * @params value:Number. Value to search in the LinkedList.
   * @returns boolean. If a node contains the value returns true, otherwise false.
   */
  searchByValue(value = null) {
    if (value === null) return false;

    let current = this.head;
    while (current) {
      if (current.value === value) return true;
      current = current.next;
    }

    return false;
  }

  /**
   * Returns an Array representing the LinkedList.
   * It does not modify the original LinkedList
   * @params
   * @returns Array: Array containing all values of the LinkedList.
   */
  toArray() {
    let el = this.head;
    let array = [];
    while (el != null) {
      array.push(el.value);
      el = el.next;
    }
    return array;
  }

  /**
   * Returns a LinkedList with reverse values.
   * It does not modify the original LinkedList
   * @params
   * @returns LinkedList: Reversed linked list
   */
  reverseLinkedList() {
    const reversed = new LinkedList();
    if (!this.head) return reversed;

    // Helper function to recursively insert in reverse order
    const insertReverse = (node) => {
      if (node.next) insertReverse(node.next);
      reversed.insert(node.value);
    };

    insertReverse(this.head);
    return reversed;
  }

  /**
   * Reverses the LinkedList values.
   * It does modify the original LinkedList.
   * @params
   * @returns Void.
   */
  reverse() {
    if (!this.head || !this.head.next) return;

    let prev = null;
    let current = this.head;
    this.tail = this.head;

    while (current) {
      const next = current.next;
      current.next = prev;
      prev = current;
      current = next;
    }

    this.head = prev;
  }
}

module.exports = LinkedList;
